/***************************************************************************
 *   Copyright (C) 2004 by Greg Grothaus                                   *
 *   ggrothau@vt.edu                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include<string> 
#include<sstream>
#include<iostream>
#include<map>
#include<vector>
#include<cstdlib>
#include<ctime>
#include<cmath>
#include<fstream>
#include<algorithm>
#include<cfloat>
#include<cstdio>
#include<cstdlib>

#include <gd.h>
#include <gdfontt.h>
#include <gdfonts.h>
#include <gdfontmb.h>
#include <gdfontl.h>
#include <gdfontg.h>

//#define FONTPATH "/src/courbd.ttf"
#define PI 3.14159

using namespace std;

struct cluster
{
	vector<int> beginX,beginY,sizeX,sizeY;
	string name,color;
};

vector<cluster> clusters;
map<string,int> clusterIndex;
map<string,map<string,bool> > classData;	//set to true if class <string, includes sample string>

vector<string> layoutX;
vector<string> layoutY;
vector<string> layoutXOrig;
vector<string> layoutYOrig;
map<string,vector<double> > allValues;

map<string,map<string,double> > level;
double offsetX,offsetY;
double width,height;
ostream *out=&cout;
FILE *gdout;

int brect[8];
int startSamples;
int nameColumn;
double sizeModifier;
double clusterWidth;
string columnReplaceFile, rowReplaceFile, classFile;
bool binaryData,drawColors,showrows,showcolumns;
bool backblue;
string imageType,blank;

gdImagePtr gdimage = 0;
string fonttmp=string(FONTPATH)+"/src/VeraBd.ttf";
char *font=(char *)fonttmp.c_str();

//START ADEEL'S CODE//
struct bicluster
{
	string name,actualColor,highlightColorName;
	vector<string> rowNames,colNames;
	int startRow,endRow,startCol,endCol,highlightColor;
};
vector<bicluster> actualBiclusters;	//will store all biclusters in file
vector<bool> highlightRows;		//will store if a row needs to be highlighted, the index will correspond to row #
vector<bool> highlightCols;		//will store if a column needs to be highlighted, the index will correspond to column #

//first int corresponds to the index number of a cluster in actualBiclusters[], first pair is start row to end row of that bicluster, second pair is start column to end column
map<int, map<pair<int,int>,pair<int,int> > > goodBiclustersToLocation;

bool namesOnRight;	//if the gene names are to be displayed on the right hand side
//mapFile format is: (polygon/rectangle):name of row/column:x1:y1:x2:y2:..., this is supposed to be read by the perl script interacting with drawlayout
ofstream mapFile;	//image map file
bool noRows;		//if there are no rows to be displayed
bool noColumns;		//if there are no columns to be displayed
//END ADEEL'S CODE//


//converts an integer to a string
template <class T> string toa(T in)
{
	stringstream ss;
	ss<<in;
	string out;
	ss>>out;
	return out;
}

template <class T> T ato(string in)
{
	stringstream ss;
	ss<<in;
	T out;
	ss>>out;
	return out;
}

//This prints a fucnction for drawing gene expression boxes
//as well as sets the font to use in the image
void printBoxFunction()
{
	//this is a filled box of size 17x17 (scaled by sizeModifier)
	*out<<"/box {"<<endl;
	*out<<"  newpath"<<endl;
	*out<<"  moveto\t\t\t% Current point is on stack"<<endl;
	*out<<"  0 "<<17*sizeModifier<<" rlineto\t\t\t% Left"<<endl;
	*out<<"  "<<17*sizeModifier<<" 0 rlineto\t\t\t% Top"<<endl;
	*out<<"  0 -"<<17*sizeModifier<<" rlineto\t\t\t% Right"<<endl;
	*out<<"  closepath\t\t\t% Bottom"<<endl;
	*out<<"  fill"<<endl;
	*out<<"} def"<<endl;
	*out<<endl;

	//sets the fonts to use in the image
	*out<<"/Courier-Bold findfont"<<endl;
	*out<<12*sizeModifier<<" scalefont"<<endl;
	*out<<"setfont"<<endl;
	*out<<endl;

}

//Calculates the offset from the origin of the actual gene expression data
//and prints the postscript header using the calculated width/height
void printHeader()
{
	//postscript:  0,0 is bottom left
	//gd: 0,0 is top left - joy
	
	//figure out the offsets:
	//Set offsetX by determining the longest gene string
	offsetX=0;
	if(!noRows)	//This will keep offsetX at 0 if no rows are to be displayed, so the image size is affected
	{
		for(int i=0;i<layoutY.size();i++)
			if(layoutY[i].length()*7*sizeModifier>offsetX)
			{
				offsetX=layoutY[i].length()*7*sizeModifier;
			}
		offsetX+=10*sizeModifier;
	}

	//Set offsetY by determining the longest sample string	
	offsetY=0;
	if(!noColumns)	//This will keep offsetY at 0 if no columns are to be displayed
	{
		for(int i=0;i<layoutX.size();i++)
			if(layoutX[i].length()*7*sizeModifier>offsetY)
				offsetY=layoutX[i].length()*7*sizeModifier;
			
		//since we are rotating by 45 degrees, we don't need the full length
		offsetY=int(double(offsetY)/double(sqrt(2.)));
		offsetY+=10*sizeModifier;
	}
	
	offsetY+=20*classData.size()*sizeModifier;
	
	width=layoutX.size()*20*sizeModifier+offsetX+offsetY;
	height=layoutY.size()*20*sizeModifier+offsetY+20*sizeModifier;
	
	if(imageType=="ps")
	{
		*out<<"%!PS-Adobe-3.0 EPSF-3.0"<<endl;
		*out<<"%%Creator: Bicluster Visualization Tool version 0.5 by Greg Grothaus"<<endl;
		*out<<"%%BoundingBox: 0 0 "<<width<<" "<<height<<endl;
		*out<<"%%LanguageLevel: 2"<<endl;
		*out<<"%%Pages: 1"<<endl;
		*out<<"%%DocumentData: Clean7Bit"<<endl;
		*out<<endl;
		printBoxFunction();
	}
	else
	{
		gdimage=gdImageCreateTrueColor(int(width),int(height));
		int white=gdImageColorAllocate (gdimage, 255, 255, 255);
		gdImageFilledRectangle (gdimage, 0, 0, gdImageSX (gdimage),
                          gdImageSY (gdimage), white);
		//gdImageColorTransparent (gdimage, white);
	
	}
}

void input(string infile,string allfile)
{
	string line;
	ifstream in(infile.c_str());
	map<string,string> crmap,rrmap;
	map<string,bool> foundcols,foundrows;

	//first the lookup table to replace gene/sample ID's, if it exists	
	if(columnReplaceFile!="")
	{
		ifstream cr(columnReplaceFile.c_str());
		while(getline(cr,line))
		{
			stringstream sscr;
			sscr<<line<<"\t";
			string a,b;
			getline(sscr,a,'\t');
			getline(sscr,b,'\t');
			crmap[a]=b;
		}
	}
	
	if(rowReplaceFile!="")
	{
		ifstream rr(rowReplaceFile.c_str());
		while(getline(rr,line))
		{
			stringstream ssrr;
			ssrr<<line<<"\t";
			string a,b;
			getline(ssrr,a,'\t');
			getline(ssrr,b,'\t');
			rrmap[a]=b;
		}
	}
			
	//now the gene/sample lists
	
	//gene list ordering
	getline(in,line);
	while(line!="")
	{
		stringstream ss;
		ss<<line;
		if(rrmap[line]!="")
			layoutY.push_back(rrmap[line]);
		else
			layoutY.push_back(line);
		layoutYOrig.push_back(line);
		foundrows[line]=true;
		line="";
		getline(in,line);
	}
	
	//gene clusters
	getline(in,line);
	while(line!="")
	{
		stringstream ss;
		ss<<line;
		cluster c;
		ss>>c.name;
		ss>>c.color;
		int size,begin;
		ss>>size;
		ss>>begin;

		if(clusters.size()>0 && clusters[clusterIndex[c.name]].name==c.name)
		{
			clusters[clusterIndex[c.name]].sizeY.push_back(size);
			clusters[clusterIndex[c.name]].beginY.push_back(begin);
		}
		else
		{
			c.sizeY.push_back(size);
			c.beginY.push_back(begin);
			clusterIndex[c.name]=clusters.size();
			clusters.push_back(c);
		}
		
		getline(in,line);
	}

	//sample list ordering
	getline(in,line);
	while(line!="")
	{
		stringstream ss;
		ss<<line;
		if(crmap[line]!="")
			layoutX.push_back(crmap[line]);
		else
			layoutX.push_back(line);
		layoutXOrig.push_back(line);
		foundcols[line]=true;
		getline(in,line);
	}

	//sample clusters
	getline(in,line);
	while(line!="")
	{
		stringstream ss;
		ss<<line;
		string name,color;
		ss>>name;
		ss>>color;
		int i=clusterIndex[name];
		
		int size,begin;
		ss>>size;
		ss>>begin;
		clusters[i].sizeX.push_back(size);
		clusters[i].beginX.push_back(begin);
		
		getline(in,line);
	}
	in.close();
	
	//output the dimensions for user interpretability
	cout<<layoutX.size()<<"x"<<layoutY.size()<<"="<<layoutX.size()*layoutY.size()<<endl;
	
	//now get the expression value data
	in.open(allfile.c_str());
	line="";
	string token,trash;
	map<int,string> colnames;
	//stores the name of each column
	
	//get/parse the header line
	getline(in,line);
	stringstream st;
	st<<line<<'\t';
	//loop through the non-sample columns
	for(int i=0;i<startSamples;i++)
		getline(st,trash,'\t');
	//get the column name data
	int idx=0;	//keep track of which column
	while(getline(st,token,'\t'))
	{
		colnames[idx++]=token;
		
		//if we have a column not in the layout
		//and showcolumns is true, add it to the layout
		if(!foundcols[token] && showcolumns)
		{
			if(crmap[token]!="")
				layoutX.push_back(crmap[token]);
			else
				layoutX.push_back(token);
			layoutXOrig.push_back(token);
		}		
	}
	

	//now loop through the data rows
	//ASSUMPTION: Only one header column
	int count=0;
	while(getline(in,line))
	{
		//ignore blank lines
		if(line=="")
			continue;
		
		stringstream ss;
		string name;
		ss<<line<<'\t';
		//get the name
		int i;
		//ignore non-data columns
		for(i=0;i<=nameColumn;i++)
			getline(ss,name,'\t');

		//if we have a row that is not in the layout
		//and showrows is true, add it to the layout		
		if(!foundrows[name] && showrows)
		{
			if(rrmap[name]!="")
				layoutY.push_back(rrmap[name]);
			else
				layoutY.push_back(name);
			layoutYOrig.push_back(name);
		}
		//if we have a row not in the layout and we don't need
		//to show it, no need to parse and store it either
		else if(!foundrows[name])
			continue;

		//get the samples
		
		//skip trash columns
		for(int i=nameColumn;i<startSamples-1;i++)
			getline(ss,trash,'\t');
			
		//now loop through all the expression values
		for(int i=0;i<idx;i++)	//idx is the number of sample columns
		{
			double value=DBL_MAX;
			getline(ss,token,'\t');
			if(token!="" && token!="NA")
			{
				stringstream ts;
				ts<<token;
				ts>>value;
			}
			level[colnames[i]][name]=value;
			allValues[name].push_back(value);
		}
	}
	//sort the values in each gene for display purposes
	if(!binaryData)
		for(map<string,vector<double> >::iterator i=allValues.begin();i!=allValues.end();i++)
			sort((*i).second.begin(),(*i).second.end());
}

void drawText()
{
	//print the gene names
	int black, red, yellow, j=0;
	if(imageType!="ps")
	{
		black=gdImageColorAllocate(gdimage,0,0,0);
		red=gdImageColorAllocate(gdimage,255,0,0);
		yellow=gdImageColorAllocate(gdimage,255,255,0);
	}

	if(!noRows)	//To check if no rows are to be displayed
	{
		for(int i=0;i<layoutY.size();i++)
		{
			if(imageType=="ps")
			{
				//first the text
				*out<<"newpath"<<endl;
				if(!namesOnRight)	//if names not on right
					*out<<offsetX-10*sizeModifier-(layoutY[i].length()*7*sizeModifier);
				else
					*out<<(width-offsetX-offsetY+sizeModifier*10);	//will display names on right hand side
				*out<<" "<<i*20*sizeModifier+2*sizeModifier<<" moveto"<<endl;
				*out<<"("<<layoutY[i]<<") show"<<endl;

				//underline
				*out<<"newpath"<<endl;
				if(!namesOnRight)	//if names not on right, fix the underline
					*out<<"0 "<<i*20*sizeModifier<<" moveto"<<endl;
				else
					*out<<(width-offsetX-offsetY+sizeModifier*10)-int(sizeModifier*10)<<" "<<i*20*sizeModifier<<" moveto"<<endl;
				*out<<sizeModifier<<" setlinewidth"<<endl;
				*out<<offsetX<<" 0 rlineto"<<endl;
				*out<<"closepath"<<endl;
				*out<<"stroke"<<endl;
			}
			else
			{
				int x,y;
				if(!namesOnRight)	//if the names are on the right
					x=int(offsetX-10*sizeModifier-(layoutY[i].length()*7*sizeModifier));
				else
					x=int(width-offsetX-offsetY+sizeModifier*10);
				y=int(height-i*20*sizeModifier-2*sizeModifier);
			
				//START ADEEL'S CODE//
				if(highlightRows[i])
				{
					if(!namesOnRight)
						gdImageFilledRectangle(gdimage, 0, int(height - (i+1)*20*sizeModifier - 1), int(offsetX), int(height-i*20*sizeModifier-1), yellow);
					else
						gdImageFilledRectangle(gdimage, x-int(sizeModifier*10), int(height - (i+1)*20*sizeModifier - 1), int(width)-int(offsetY), int(height-i*20*sizeModifier-1), yellow);
				}
				//END ADEEL'S CODE//

				gdImageStringFT(gdimage,&brect[0],black,font,7*sizeModifier,0.0,x,y,(char *)layoutY[i].c_str());

				y=int(height-i*20*sizeModifier-1);
				gdImageSetThickness(gdimage, 1);
				if(!namesOnRight)
				{
					mapFile << "r" << ":" << layoutY[i] << ":" << 0 << ":" << int(height - (i+1)*20*sizeModifier - 1) << ":" << int(offsetX) << ":" << int(height-i*20*sizeModifier-1) << endl;
					gdImageLine(gdimage,0,y,int(offsetX),y,black);
				}
				else
				{
					mapFile << "r" << ":" << layoutY[i] << ":" << x-int(sizeModifier*10) << ":" << int(height - (i+1)*20*sizeModifier - 1) << ":" << int(width)-int(offsetY) << ":" << int(height-i*20*sizeModifier-1) << endl;
					gdImageLine(gdimage,x-int(sizeModifier*10),y,int(width)-int(offsetY),y,black);
				}
			}
		}
	}
	
	//treat classes the same way as genes
	//print the gene names
	for(map<string,map<string,bool> >::iterator i=classData.begin();i!=classData.end();i++)
	{
		if(!noRows)
		{
			if(imageType=="ps")
			{
				//text
				*out<<"newpath"<<endl;
				if(!namesOnRight)
					*out<<offsetX-10*sizeModifier-(i->first.length()*7*sizeModifier);
				else
					*out<<(width-offsetX-offsetY+sizeModifier*10);
				*out<<" "<<(layoutY.size()+j)*20*sizeModifier+2*sizeModifier<<" moveto"<<endl;
				*out<<"("<<i->first<<") show"<<endl;

				//underline
				*out<<"newpath"<<endl;
				if(!namesOnRight)
					*out<<"0 "<<(layoutY.size()+j)*20*sizeModifier<<" moveto"<<endl;
				else
					*out<<(width-offsetX-offsetY+sizeModifier*10)-int(sizeModifier*10)<<" "<< (layoutY.size()+j)*20*sizeModifier<<" moveto"<<endl;
				*out<<sizeModifier<<" setlinewidth"<<endl;
				*out<<offsetX<<" 0 rlineto"<<endl;
				*out<<"closepath"<<endl;
				*out<<"stroke"<<endl;
			}
			else
			{
				int x, y;
				if(!namesOnRight)
					x=int(offsetX-10*sizeModifier-(i->first.length()*7*sizeModifier));
				else
					x=int(width-offsetX-offsetY+sizeModifier*10);
				y=int(height-(layoutY.size()+j)*20*sizeModifier-2*sizeModifier);

				gdImageStringFT(gdimage,&brect[0],black,font,7*sizeModifier,0.0,x,y,(char *)i->first.c_str());
				y=int(height-(layoutY.size()+j)*20*sizeModifier-1);
				gdImageSetThickness(gdimage, 1);
					//gdImageLine(gdimage,0,y,int(offsetX),y,black);
				if(!namesOnRight)
					gdImageLine(gdimage,0,y,int(offsetX),y,black);
				else
					gdImageLine(gdimage,x-int(sizeModifier*10),y,int(width)-int(offsetY),y,black);
			}
		}
		j++;
	}

	if(!noColumns)
	{
		//calculate the actual length of the longest sample name
		int linelength=int(double(offsetY)*sqrt(2.)+1);
	
		double lys=layoutY.size()+classData.size();
	
		//loop through the sample names, and print them, rotated
		for(int i=0;i<layoutX.size();i++)
		{
			if(namesOnRight) offsetX=0;
			if(imageType=="ps")
			{
				//first the word
				*out<<"newpath"<<endl;
				*out<<"gsave"<<endl;
				*out<<offsetX+i*20*sizeModifier+18*sizeModifier<<" "
				    <<lys*20*sizeModifier+2*sizeModifier<<" moveto"<<endl;
				*out<<"45 rotate"<<endl;
				*out<<"("<<layoutX[i]<<") show"<<endl;
				*out<<"grestore"<<endl;
			
				//then the underlining
				*out<<"gsave"<<endl;
				*out<<"newpath"<<endl;
				*out<<offsetX+i*20*sizeModifier+20*sizeModifier<<" "
				    <<lys*20*sizeModifier<<" moveto"<<endl;
				*out<<"45 rotate"<<endl;
				*out<<sizeModifier<<" setlinewidth"<<endl;
				*out<<linelength<<" 0 rlineto"<<endl;
				*out<<"closepath"<<endl;
				*out<<"stroke"<<endl;
				*out<<"grestore"<<endl;
			}
			else
			{
				gdPoint points[4];
	
				int x=int(offsetX+i*20*sizeModifier+20*sizeModifier);
				int y=int(height-lys*20*sizeModifier);

				//START ADEEL'S CODE//
				//calculate the points for the polygon needed to be highlighted
				points[0].x=x;
				points[0].y=int(height-(layoutY.size()+j)*20*sizeModifier);
				points[1].x=int(offsetX+(i-1)*20*sizeModifier+20*sizeModifier);
				points[1].y=int(height-(layoutY.size()+j)*20*sizeModifier);
				points[2].x=int(int(offsetX+(i-1)*20*sizeModifier+20*sizeModifier)+offsetY);
				points[2].y=int(int(height-(layoutY.size()+j)*20*sizeModifier)-offsetY);
				points[3].x=int(x+offsetY);
				points[3].y=int(int(height-(layoutY.size()+j)*20*sizeModifier)-offsetY);

				if(highlightCols[i])	//if selected by pickGoodBiclusters()
					gdImageFilledPolygon(gdimage,points,4,yellow);

				mapFile << "p" << ":" << layoutX[i] << ":" << points[0].x << ":" << points[0].y << ":" << points[1].x << ":" << points[1].y << ":" << points[2].x << ":" << points[2].y << ":" << points[3].x << ":" << points[3].y << endl;
				//END ADEEL'S CODE//

				gdImageStringFT(gdimage,&brect[0],black,font,7*sizeModifier,PI/4,x,y,(char *)layoutX[i].c_str());

				y=int(height-(layoutY.size()+j)*20*sizeModifier-1);
				gdImageSetThickness(gdimage, 1);
				gdImageLine(gdimage,x,y,int(x+offsetY),int(y-offsetY),black);
			}	
		}
	}
}

double getRed(string x, string y)
{
	if(binaryData)
		return level[x][y];
	
	double exp=level[x][y];
	if(exp==DBL_MAX)
		return exp;
	int a=0;
	int max=allValues[y].size();
	for(int i=max-1;i>-1;i--)
	{
		if(allValues[y][i]==DBL_MAX)
			max--;
		else
			break;
	}
	int b=max;
	double percent=double(b)/100.;
	while((b-a)>percent && (b-a)>1)
	{
		int c=(a+b)/2;
		if(allValues[y][c]>exp)
			b=c;
		else if(allValues[y][c]<exp)
			a=c;
		else return ( (1 - double(c)/double(max)) );
	}
	return (1 -( double(a)/double(max)) );
	
}

void drawBoxes()
{
	if(namesOnRight) offsetX=0;
	//increase the offsets slightly to put them past the edges of the words
	offsetX+=3*sizeModifier;
	offsetY+=3*sizeModifier;
	int blue,black,grey,grey2,green,purple,red;
	if(imageType!="ps")
	{
		grey=gdImageColorAllocate(gdimage,220,220,220);
		purple=gdImageColorAllocate(gdimage,255,0,255);
		grey2=gdImageColorAllocate(gdimage,128,128,128);
		green=gdImageColorAllocate(gdimage,0,255,0);
		black=gdImageColorAllocate(gdimage,0,0,0);
		blue=gdImageColorAllocate(gdimage,0,0,255);
		red=gdImageColorAllocate(gdimage,255,0,0);
	}

	//double loop through the classes, draw purple
	int idx=0;
	for(map<string,map<string,bool> >::iterator i=classData.begin();i!=classData.end();i++)
	{
		bool drawThis=false;
		for(int j=0;j<layoutX.size();j++)
			if(i->second[layoutX[j]])
				drawThis=true;
		if(drawThis)
		{
			for(int j=0;j<layoutX.size();j++)
			{
				if(imageType=="ps")
				{
					if(i->second[layoutX[j]])
						*out<<"1.0 0.0 1.0 setrgbcolor"<<endl;
					else
						*out<<"0.9 0.9 0.9 setrgbcolor"<<endl;
					//now that we have the color, draw the box
					//using the predefined \box ps function
					*out<<offsetX+(20*j*sizeModifier)<<" "
						<<(20*(layoutY.size()+idx)*sizeModifier)
						<<" box"<<endl;
				}
				else
				{
					int color;
					if(i->second[layoutX[j]])
						color=purple;
					else
						color=grey;
					int x=int(offsetX+(20*j*sizeModifier));
					int y=int(height-(20*(layoutY.size()+idx)*sizeModifier)-19*sizeModifier);

					gdImageFilledRectangle(gdimage,x,y,x+int(17*sizeModifier),y+int(17*sizeModifier),color);

				}
			}
			idx++;
		}
	}
		
	//double loop through the matrix, draw the heat diagram
	for(int i=0;i<layoutX.size();i++)
	{
		for(int j=0;j<layoutY.size();j++)
		{
			int color;
			//get an idea of how red this should be
			double red=getRed(layoutXOrig[i],layoutYOrig[j]);
			
			//error checking, red should be between 0 and 1
			//otherwise output grey
			//if red is DBL_MAX, it is a missing value
			if(red<0 || red>1 || red==DBL_MAX)
				if(imageType=="ps")
					*out<<"0.5 0.5 0.5 setrgbcolor"<<endl;
				else
					color=grey2;
			else
			{
				if(binaryData)
				{
					//green versus grey
					if(red==1)
						if(imageType=="ps")
							*out<<"0.0 1.0 0.0 setrgbcolor"<<endl;
						else
							color=green;
					else
						if(imageType=="ps")
							*out<<"0.9 0.9 0.9 setrgbcolor"<<endl;
						else
							color=grey;
				}
				else
				{
					//real valued data
					//red=overexpressed
					//green=underexpressed
					//black=mean expression
					if(imageType=="ps")
					{
						if(red<0.5)
							*out<<2*(0.5-red)<<" 0.0 0.0 setrgbcolor"<<endl;
						else if(red>=0.5)
							*out<<"0.0 "<<2*(red-0.5)<<" 0.0 setrgbcolor"<<endl;
					}
					
					else
					{
						if(red<0.5)
							color=gdImageColorAllocate(gdimage,int(510*(0.5-red)),0,0);
						else if(red>=0.5)
							color=gdImageColorAllocate(gdimage,0,int(510*(red-0.5)),0);
					}
				}
			}
			//now that we have the color, draw the box
			if(imageType=="ps")
				*out<<offsetX+(20*i*sizeModifier)<<" "<<(20*j*sizeModifier)
					<<" box"<<endl;
			else
			{
				int x=int(offsetX+(20*i*sizeModifier));
				int y=int(height-(20*j*sizeModifier)-19*sizeModifier);
				gdImageFilledRectangle(gdimage,x,y,x+int(17*sizeModifier),y+int(17*sizeModifier),color);
			}
		}
	}
		
	vector <int> highlight_x1, highlight_y1, highlight_x2, highlight_y2, highlight_colors;	//will store the x and y values and the colors for highlight bicluster boxes

	//now overlay the bicluster boxes
	for(int i=0;i<clusters.size();i++)
	{
		int color;
		cluster c=clusters[i];
		if(drawColors)
		{
			if(imageType=="ps")
			{
				if(c.color=="black")
					*out<<"0.0 0.0 0.0 setrgbcolor"<<endl;
				else if(c.color=="red")
					*out<<"1.0 0.0 0.0 setrgbcolor"<<endl;
				else if(c.color=="blue")
					*out<<"0.0 0.0 1.0 setrgbcolor"<<endl;
				else if(c.color=="green")
					*out<<"0.0 1.0 0.0 setrgbcolor"<<endl;
				else
					*out<<"0.0 0.0 1.0 setrgbcolor"<<endl;
			}
			else
			{
				if(c.color=="black")
					color=black;
				else if(c.color=="red")
					color=red;
				else if(c.color=="blue")
					color=blue;
				else if(c.color=="green")
					color=green;
				else
					color=blue;
			}
		}
		else
		{
			if(imageType=="ps")
			{
				//draw xMotifs that repeat in red	
				if(c.sizeX.size()>1 || c.sizeY.size()>1)
					*out<<"1.0 0.0 0.0 setrgbcolor"<<endl;
				//draw other xMotifs in blue
				else
					*out<<"0.0 0.0 1.0 setrgbcolor"<<endl;
			}
			else
			{
				//draw xMotifs that repeat in red	
				if(c.sizeX.size()>1 || c.sizeY.size()>1)
					color=red;
				//draw other xMotifs in blue
				else
					color=blue;
			}
		}

		if(sizeModifier>0.5 && imageType!="ps")
			gdImageSetThickness(gdimage, 2);
		//loop through all instances of this xMotif
		//and draw it where it stands
		
		for(int j=0;j<c.sizeX.size();j++)
		{
			for(int k=0;k<c.sizeY.size();k++)
			{
				if(imageType=="ps")
				{
					*out<<"%"<<c.name<<endl;
					*out<<"newpath"<<endl;
					*out<<clusterWidth<<" setlinewidth"<<endl;
					*out<<offsetX+(20*c.beginX[j]*sizeModifier)-
						2*sizeModifier<<" "<<(20*c.beginY[k]*sizeModifier)-
						2*sizeModifier<<" moveto"<<endl;
					*out<<"0 "<<20*c.sizeY[k]*sizeModifier<<" rlineto"<<endl;
					*out<<20*c.sizeX[j]*sizeModifier<<" 0 rlineto"<<endl;
					*out<<"0 "<<-20*c.sizeY[k]*sizeModifier<<" rlineto"<<endl;
				
					*out<<"closepath"<<endl;
					*out<<"stroke"<<endl<<endl;
				}
				else
				{
					bool found=false;

					int x1=int(offsetX+(20*c.beginX[j]*sizeModifier)-2*sizeModifier+1);
					int y1=int(height-(20*(c.beginY[k]+c.sizeY[k]-1)*sizeModifier)-21*sizeModifier+1);
					int x2=int(offsetX+(20*(c.beginX[j]+c.sizeX[k])*sizeModifier)-2*sizeModifier);
					int y2=int(height-(20*(c.beginY[k]-1)*sizeModifier)-21*sizeModifier);

					//START ADEEL'S CODE//
					//this will look for the goodBiclusters that need to be highlighted
					for(map<int, map<pair<int,int>,pair<int,int> > >::iterator i=goodBiclustersToLocation.begin(); i!=goodBiclustersToLocation.end(); i++)
					{
						if(actualBiclusters[(*i).first].name==c.name)
						{
							highlight_x1.push_back(x1);
							highlight_y1.push_back(y1);
							highlight_x2.push_back(x2);
							highlight_y2.push_back(y2);
							highlight_colors.push_back(actualBiclusters[(*i).first].highlightColor);

							found=true;
							break;
						}
					}
					//END ADEEL'S CODE//

					if(!found)
						gdImageRectangle(gdimage,x1,y1,x2,y2,color);

				}
				
			}
		}
	}

	//START ADEEL'S CODE//
	if(imageType!="ps")	//This will draw all the highlighted bicluster boxes at the end so they are not overlapped
	{
		gdImageSetThickness(gdimage, 4);
		for(int i=0; i<highlight_x1.size(); i++)
			gdImageRectangle(gdimage,highlight_x1[i],highlight_y1[i],highlight_x2[i],highlight_y2[i],highlight_colors[i]);
		gdImageSetThickness(gdimage, 1);
	}
	//END ADEEL'S CODE//

}

void readClasses(string file)
{
	ifstream in(file.c_str());
	
	//loop through the class file
	string line;
	while(getline(in,line))
	{
		stringstream ss;
		ss<<line;
		string sample;
		getline(ss,sample,'\t');
		bool isLayout=false;
		for(int i=0;i<layoutX.size();i++)
			if(layoutX[i]==sample)
				isLayout=true;
		if(!isLayout)
			continue;
		while(getline(ss,line,'\t'))
			classData[line][sample]=true;
	}
	
	in.close();
}			

//START ADEEL'S CODE//
//this will set up the biclusters so we can search through them
void setupBiclusterData()
{
	for(int i=0;i<clusters.size();i++)
	{
		int begin, end;
		bicluster b;

		b.name=clusters[i].name;
		b.actualColor=clusters[i].color;
		b.highlightColorName="yellow";  //THIS IS WHERE YOU CAN DECIDE WHAT COLOR YOU WANT, based on if elses?
		b.highlightColor=gdImageColorAllocate(gdimage,255,255,0);

		//start rows and columns to their end in layoutX or layoutY;
		b.startCol=clusters[i].beginY[0];
		b.endCol=clusters[i].beginY[0]+clusters[i].sizeY[0];
		b.startRow=clusters[i].beginX[0];
		b.endRow=clusters[i].beginX[0]+clusters[i].sizeX[0];

		//then the column names
		for (int k=0; k<clusters[i].beginX.size(); k++)
			for(int l=clusters[i].beginX[k]; l<clusters[i].beginX[k]+clusters[i].sizeX[k]; l++)
				b.colNames.push_back(layoutX[l]);

		//row names
		for (int k=0; k<clusters[i].beginY.size(); k++)
			for(int l=clusters[i].beginY[k]; l<clusters[i].beginY[k]+clusters[i].sizeY[k]; l++)
				b.rowNames.push_back(layoutY[l]);

		actualBiclusters.push_back(b);
	}


}

//this will pick search terms that are in the biclusters
void pickGoodBiclusters(vector <string> searchTerms, bool searchInclusive, double goodPercentage)
{
	int i, j, k, count;
	vector<int> tmpRow, tmpCol;	//these will hold rows and columns that need to be highlighted
 
	if(!searchInclusive)	//if the search is "or", meaning if any terms are found at all
		goodPercentage=0;

	for(i=0; i<actualBiclusters.size(); i++)
	{
		tmpRow.clear();
		tmpCol.clear();

		count=0;
		//go through rows and find any matches with search terms
		for(j=0; j<actualBiclusters[i].rowNames.size(); j++)
		{
			for(k=0; k<searchTerms.size(); k++)
				if(actualBiclusters[i].rowNames[j]==searchTerms[k])
				{
					tmpRow.push_back(actualBiclusters[i].startCol+j);
					count++;
				}
		}

		//go through columns and find any matches with search terms
		for(j=0; j<actualBiclusters[i].colNames.size(); j++)
		{
			for(k=0; k<searchTerms.size(); k++)
				if(actualBiclusters[i].colNames[j]==searchTerms[k])
				{
					tmpCol.push_back(actualBiclusters[i].startRow+j);
					count++;
				}
		}

		//basically use the percentage to see what number of terms are in this bicluster and if its acceptable
		if((double(count)/double(searchTerms.size()) >= goodPercentage) && (count!=0))
		{
			pair<int,int>a,b;
			for(j=0; j<tmpRow.size(); j++)
				highlightRows[tmpRow[j]]=true;
			for(j=0; j<tmpCol.size(); j++)
				highlightCols[tmpCol[j]]=true;

			a.first=actualBiclusters[i].startRow;	//start row to end row to put in map
			a.second=actualBiclusters[i].endRow;
			b.first=actualBiclusters[i].startCol;	//start column to end column
			b.second=actualBiclusters[i].endCol;
			goodBiclustersToLocation[i][a]=b;	//setup location
		}
	}
}

//for chopping the image, it will just dump out the good biclusters in the proper format to be read by layout
//so the process is that perl script calls drawlayout with chop, which doesnt draw an image, then calls layout
//on the newly generated bicluster file and then runs drawlayout on the new layout file with selected biclusters
void writeGoodBiclusters(string goodBiclusterFileName)
{
	int j;
	bicluster b;
	ofstream out;

	//(*i).second.first.first is start row, .second.first.second is end row
	map<int, map<pair<int,int>,pair<int,int> > >::iterator i;

	out.open(goodBiclusterFileName.c_str(), ofstream::out | ofstream::app);

	for(i=goodBiclustersToLocation.begin(); i!=goodBiclustersToLocation.end(); i++)
	{
		b=actualBiclusters[(*i).first];
		out << b.name << "_" << b.colNames.size() << "_" << b.rowNames.size() << "\t" << b.actualColor;
		for(j=0; j<b.colNames.size(); j++)
			out << "\t" << b.colNames[j];
		for(j=0; j<b.rowNames.size(); j++)
			out << "\t" << b.rowNames[j];
		out << endl;
	}

	out.close();
}
//END ADEEL'S CODE//


/* Main Function
 * Parses the command line parameters
 */
int main(int argc,char** argv)
{
	bool doSearch, chop, searchInclusive;
	string mapFileName, goodBiclusterFileName;
	double goodPercentage;
	vector<string> searchTerms;

	//set default values for global parameters
	startSamples=1;
	nameColumn=0;
	sizeModifier=1.;
	clusterWidth=3.;
	columnReplaceFile=rowReplaceFile=classFile="";
	binaryData=showcolumns=showrows=drawColors=false;
	string inputfile,allfile,outfile;
	imageType="ps";
	namesOnRight=false;
	int searchCount=0;
	chop=false;
	doSearch=false;
	noRows=false;
	noColumns=false;

	//loop through the parameter list
	for(int i=1;i<argc;i++)
	{
		//output file
		if(argv[i]==string("-o"))
			outfile=argv[++i];
		//output file type
		else if(argv[i]==string("-t"))
		{
			imageType=argv[++i];
			if(imageType!="gif" && imageType!="ps" && imageType!="png")
			{
				cerr<<"Image Type "<<imageType<<" is unrecognized!"<<endl;
				return 1;
			}	
		}

		//expression value file
		else if(argv[i]==string("-a"))
			allfile=argv[++i];
		//The column where the first sample begins
		//0 based indexing
		else if(argv[i]==string("-startSamples"))
			startSamples=ato<int>(argv[++i]);
		//The column containing gene names
		//0 based indexing
		else if(argv[i]==string("-nameColumn"))
			nameColumn=ato<int>(argv[++i]);
		//resizes the entire image by some value
		//A value of 0.5 means that the image will
		//be 50% the original height and width, or
		//25% the original total size
		else if(argv[i]==string("-sizeModifier"))
			sizeModifier=ato<double>(argv[++i]);
		//Same as sizeModifier, but applies to the width
		//of the xMotif box lines
		else if(argv[i]==string("-width"))
			clusterWidth=ato<double>(argv[++i]);
		//specify a mapping from sampleID's to sample names
		else if(argv[i]==string("-cr"))
			columnReplaceFile=argv[++i];
		//specify a mapping from geneID's to gene names		
		else if(argv[i]==string("-rr"))
			rowReplaceFile=argv[++i];
		//If all the data values are 0/1, use this flag
		//Color interpolation will not be used, instead
		//on/off appropriate colors will
		else if(argv[i]==string("-binary"))
			binaryData=true;
		//uses the color field in the bicluster specification
		//to determine what color to draw the cluster in
		//currently only recognizes (black,red,green,blue)
		//default is blue
		else if(argv[i]==string("-colors"))
			drawColors=true;
		//shows all columns in the data, even if no bicluster
		else if(argv[i]==string("-showColumns"))
			showcolumns=true;
		//shows all rows in the data, even if no bicluster
		else if(argv[i]==string("-showRows"))
			showrows=true;
		//shows all rows in the data, even if no bicluster
		else if(argv[i]==string("-classes"))
			classFile=argv[++i];
		else if(argv[i]==string("-search"))
			doSearch=true;
		else if(argv[i]==string("-r"))
			namesOnRight=true;
		else if(argv[i]==string("-chop"))
		{
			goodBiclusterFileName=argv[++i];
			chop=true;
		}
		else if(argv[i]==string("-noRows"))
			noRows=true;
		else if(argv[i]==string("-noColumns"))
			noColumns=true;
		//show the help display
		else if(argv[i]==string("--help") || argv[i]==string("-help") || argv[i]==string("-h"))
		{
			cout<<"Bicluster Layout Drawing Tool"<<endl;
			cout<<"Usage: drawlayout [<OPTIONS>] <bicluster layout file>"<<endl;
			cout<<" -o <file>              Specify Output File (postscript)"<<endl;
			cout<<" -a <file>              Specify Expression Values input file"<<endl;
			cout<<" -h                     This Screen"<<endl;
			cout<<"--------------Options------------------"<<endl;
			cout<<" -startSamples <int>    Specify the Column of first sample"<<endl;
			cout<<" -nameColumn   <int>    Specify the Column of the gene names"<<endl;
			cout<<" -sizeModifier <double> Multiply all dimensions by this value"<<endl;
			cout<<" -width        <double> Multiply bicluster line width"<<endl;
			cout<<" -cr           <file>   Column ID replacement mapping"<<endl;
			cout<<" -rr           <file>   Row ID replacement mapping"<<endl;
			cout<<" -binary                Specifies that expression data is binary"<<endl;
			cout<<" -colors                Use the colors in the bicluster file"<<endl;
			cout<<" -showColumns           Show all columns in the expression file"<<endl;
			cout<<" -showRows              Show all rows in the expression file"<<endl;
			cout<<" -classes      <file>   Define classes for the samples"<<endl;
			cout<<" -t            <string> Output image format(ps,png,gif)"<<endl;
			cout<<" -search                Highlight gene names or sample names"<<endl;
			cout<<" -r                     Put gene names on the right side"<<endl;
			cout<<" -chop         <file>   Will output a new bicluster file with only good biclusters"<<endl;
			cout<<" -noRows                Display no rows"<<endl;
			cout<<" -noColumns             Display no columns"<<endl;
			return 0;
		}
		//default parameter (bicluster input file)
		else if(argv[i-1][0]!='-' && argv[i][0]!='-')
			inputfile=argv[i];
		//handles undefined flags
		else
		{
			cerr<<argv[i]<<" is unrecognized!\n";
			return 1;
		}
	}
	//check that there is a bicluster file and expression file
	//no output is needed as by default, output goes to standard out
	if(inputfile=="" || allfile=="")
	{
		cerr<<"Error, require bicluster and expression input files.\n -h for Help\n";
		exit(1);
	}

	input(inputfile,allfile);

	if(classFile!="")
		readClasses(classFile);

	//use the default postscript generator encoded in this software
	if(imageType=="ps")
		out = new ofstream(outfile.c_str());
	else
		gdout = fopen(outfile.c_str(),"wb");


	printHeader();


	//START ADEEL'S CODE//
	//initialize these to false so no rows/cols are to be highlighted initially
	for(int i=0; i<layoutX.size(); i++)
		highlightCols.push_back(false);
	for(int i=0; i<layoutY.size(); i++)
		highlightRows.push_back(false);

	if(chop && !doSearch)
	{
		cerr << "You cannot chop without doing search\n";
		exit(1);
	}

	//this reads from stdin first if its "and" or an "or" search, if its an "and" it reads the percentage specified
	//by the user through the perl script drawlayout interacts with. Then it reads the number of search terms
 	//and then reads the actual search terms
	if(doSearch)
	{
		int ii; string tmp;

		cin >> tmp;
		if(tmp=="and") searchInclusive=true;
		else if(tmp=="or") searchInclusive=false;
		else { cerr << "Error, wrong input for searching!" << endl; exit(1); }
		if(tmp=="and") cin >> goodPercentage;
		cin >> searchCount;
		for (ii=0; ii<searchCount; ii++)
		{
			cin >> tmp;
			searchTerms.push_back(tmp);
		}
	
		setupBiclusterData();	//call to setup bicluster data first
		pickGoodBiclusters(searchTerms, searchInclusive, goodPercentage);	//call to pick good biclusters

		//just output the good biclusters found by pickGoodBiclusters and exit, dont draw any image
		if(chop)
		{
			writeGoodBiclusters(goodBiclusterFileName);
			exit(0);
		}
	}

	mapFileName=outfile+".map";
	mapFile.open(mapFileName.c_str(), ofstream::out | ofstream::app);
	//END ADEEL'S CODE//


	drawText();
	drawBoxes();

	if(imageType!="ps")
	{
		gdImagePng(gdimage, gdout);
		fclose(gdout);
		gdImageDestroy(gdimage);
	}

	mapFile.close();
	
}
